// PV112 2017, test for lesson 4
// Libraries needed: GL, GLEW, GLUT, GLM, DevIL

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <math.h>

#define GLEW_STATIC
#include <GL/glew.h>

#if defined(_WIN32)
#pragma comment(lib, "glew32s.lib")
// On Windows, we use Unicode dll of DevIL
// That also means we need to use wide strings
#ifndef _UNICODE
#define _UNICODE
#include <IL/il.h>
#undef _UNICODE
#else
#include <IL/il.h>
#endif
#pragma comment(lib, "DevIL.lib") // Link with DevIL library
typedef wchar_t maybewchar;
#define MAYBEWIDE(s) L##s
#define PRINTFSTR "%S"

#else // On Linux, we need regular (not wide) strings
#include <IL/il.h>
typedef char maybewchar;
#define MAYBEWIDE(s) s
#define PRINTFSTR "%s"
#endif

// Include FreeGLUT
#include <GL/freeglut.h>

// Include the most important GLM functions
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

// Shader program and its uniforms
GLuint program;
GLint PVM_matrix_loc;
GLint my_tex_loc;

GLuint program2;
GLint PVM_matrix_loc2;
GLint my_tex1_loc2;
GLint my_tex2_loc2;
GLint my_alpha_loc2;

PV112Geometry my_cube;
PV112Camera my_camera;

GLuint lenna_tex;
GLuint lenna512_tex;
GLuint lenna511_tex;
GLuint lenna510_tex;
GLuint lenna509_tex;
GLuint lenna508_tex;
GLuint wood_tex;
GLuint alpha_circle_tex;

float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
        exit(0);
        break;
    case 'l':
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glutPostRedisplay();
        break;
    case 'f':
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glutPostRedisplay();
        break;
    case 't':
        glutFullScreenToggle();
        break;
    }
}

// Called when the user presses a mouse button
void mouse_button_changed(int button, int state, int x, int y)
{
    my_camera.OnMouseButtonChanged(button, state, x, y);
}

// Called when the user moves the mouse (while a mouse button is being pressed)
void mouse_moved(int x, int y)
{
    my_camera.OnMouseMoved(x, y);
}

// Loads a texture from file and calls glTexImage2D to set its data.
// Returns true on success or false on failure.
bool LoadAndSetTexture(const maybewchar *filename, GLenum target)
{
    // NOTE: When using other library than DevIL, rewrite this code

    // Create IL image
    ILuint IL_tex;
    ilGenImages(1, &IL_tex);

    ilBindImage(IL_tex);

    // Solve upside down textures
    // When the image is upside down, find a way to load it correctly
    ilEnable(IL_ORIGIN_SET);
    ilOriginFunc(IL_ORIGIN_LOWER_LEFT);

    // Load IL image
    ILboolean success = ilLoadImage(filename);
    if (!success)
    {
        ilBindImage(0);
        ilDeleteImages(1, &IL_tex);
        printf("Couldn't load texture: " PRINTFSTR "\n", filename);
        return false;
    }

    // Get IL image parameters
    int img_width = ilGetInteger(IL_IMAGE_WIDTH);
    int img_height = ilGetInteger(IL_IMAGE_HEIGHT);
    int img_format = ilGetInteger(IL_IMAGE_FORMAT);
    int img_type = ilGetInteger(IL_IMAGE_TYPE);

    // Choose internal format, format, and type for glTexImage2D
    // When colors are incorrect, find proper 'format'
    // When alpha is incorrect, find proper 'format' and 'internal_format'
    GLint internal_format = 0;
    GLenum format = 0;
    GLenum type = img_type; // IL constants matches GL constants
    switch (img_format) {
    case IL_RGB:  internal_format = GL_RGB;  format = GL_RGB;  break;
    case IL_RGBA: internal_format = GL_RGBA; format = GL_RGBA; break;
    case IL_BGR:  internal_format = GL_RGB;  format = GL_BGR;  break;
    case IL_BGRA: internal_format = GL_RGBA; format = GL_BGRA; break;
    case IL_COLOR_INDEX:
    case IL_ALPHA:
    case IL_LUMINANCE:
    case IL_LUMINANCE_ALPHA:
        // Unsupported format
        ilBindImage(0);
        ilDeleteImages(1, &IL_tex);
        printf("Texture " PRINTFSTR " has unsupported format\n", filename);
        return false;
    }

    // Set the data to OpenGL (assumes texture object is already bound)
    // When images of some sizes (mostly not multiples of 4) are loaded
    // incorrectly, try different values for GL_UNPACK_ALIGNMENT or other
    // parameters of glPixelStorei, see the documentation of this function.
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(target, 0, internal_format, img_width, img_height, 0, format,
            type, ilGetData());

    // Unset and delete IL texture
    ilBindImage(0);
    ilDeleteImages(1, &IL_tex);

    return true;
}

GLuint CreateAndLoadTexture(const maybewchar *filename)
{
    // Create OpenGL texture object
    GLuint tex_obj;
    glGenTextures(1, &tex_obj);
    glBindTexture(GL_TEXTURE_2D, tex_obj);

    // Load the data into OpenGL texture object
    if (!LoadAndSetTexture(filename, GL_TEXTURE_2D))
    {
        glBindTexture(GL_TEXTURE_2D, 0);
        glDeleteTextures(1, &tex_obj);
        return 0;
    }
    glBindTexture(GL_TEXTURE_2D, 0);

    return tex_obj;
}

GLuint CreateAndLoadTextureCube(
    const maybewchar *filename_px, const maybewchar *filename_nx,
    const maybewchar *filename_py, const maybewchar *filename_ny,
    const maybewchar *filename_pz, const maybewchar *filename_nz)
{
    // Create OpenGL texture object
    GLuint tex_obj;
    glGenTextures(1, &tex_obj);
    glBindTexture(GL_TEXTURE_CUBE_MAP, tex_obj);

    // Load the data into OpenGL texture object
    if (
        !LoadAndSetTexture(filename_px, GL_TEXTURE_CUBE_MAP_POSITIVE_X) ||
        !LoadAndSetTexture(filename_nx, GL_TEXTURE_CUBE_MAP_NEGATIVE_X) ||
        !LoadAndSetTexture(filename_py, GL_TEXTURE_CUBE_MAP_POSITIVE_Y) ||
        !LoadAndSetTexture(filename_ny, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y) ||
        !LoadAndSetTexture(filename_pz, GL_TEXTURE_CUBE_MAP_POSITIVE_Z) ||
        !LoadAndSetTexture(filename_nz, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z))
    {
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
        glDeleteTextures(1, &tex_obj);
        return 0;
    }
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

    return tex_obj;
}

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    glEnable(GL_DEPTH_TEST);

    int position_loc = 0;
    int normal_loc = 1;
    int tex_coord_loc = 2;

    // Create shader programs, force location
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl",
        position_loc, "position",
        normal_loc, "normal",
        tex_coord_loc, "tex_coord");

    program2 = CreateAndLinkProgram("vertex.glsl", "fragment2.glsl",
        position_loc, "position",
        normal_loc, "normal",
        tex_coord_loc, "tex_coord");

    if (0 == program)
        WaitForEnterAndExit();

    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");
    my_tex_loc = glGetUniformLocation(program, "my_tex");

    PVM_matrix_loc2 = glGetUniformLocation(program2, "PVM_matrix");
    my_tex1_loc2 = glGetUniformLocation(program2, "my_tex1");
    my_tex2_loc2 = glGetUniformLocation(program2, "my_tex2");
    my_alpha_loc2 = glGetUniformLocation(program2, "my_alpha");

    my_cube = CreateCube(position_loc, normal_loc, tex_coord_loc);

    // Load textures
    lenna_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna.png"));
    glBindTexture(GL_TEXTURE_2D, lenna_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    lenna512_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna512.bmp"));
    glBindTexture(GL_TEXTURE_2D, lenna512_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    lenna511_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna511.bmp"));
    glBindTexture(GL_TEXTURE_2D, lenna511_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    lenna510_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna510.bmp"));
    glBindTexture(GL_TEXTURE_2D, lenna510_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    lenna509_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna509.bmp"));
    glBindTexture(GL_TEXTURE_2D, lenna509_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    lenna508_tex = CreateAndLoadTexture(MAYBEWIDE("Lenna508.bmp"));
    glBindTexture(GL_TEXTURE_2D, lenna508_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    wood_tex = CreateAndLoadTexture(MAYBEWIDE("wood.jpg"));
    glBindTexture(GL_TEXTURE_2D, wood_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    alpha_circle_tex = CreateAndLoadTexture(MAYBEWIDE("alpha_circle.png"));
    glBindTexture(GL_TEXTURE_2D, alpha_circle_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;
    glm::mat3 normal_matrix;

    projection_matrix = glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 100.0f);
    view_matrix = glm::lookAt(my_camera.GetEyePosition(), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));

    glUseProgram(program);
    glUniform1i(my_tex_loc, 0);
    glActiveTexture(GL_TEXTURE0);

    glBindVertexArray(my_cube.VAO);

    glBindTexture(GL_TEXTURE_2D, lenna_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 6.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, lenna512_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(-6.0f, 3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, lenna511_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(-3.0f, 3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, lenna510_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, lenna509_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(3.0f, 3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, lenna508_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(6.0f, 3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindTexture(GL_TEXTURE_2D, wood_tex);
    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glUseProgram(program2);
    glUniform1i(my_tex1_loc2, 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, wood_tex);
    glUniform1i(my_tex2_loc2, 1);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, lenna_tex);
    glUniform1i(my_alpha_loc2, 2);
    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, alpha_circle_tex);

    model_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -3.0f, 0.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindVertexArray(0);
    glUseProgram(0);

    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type)
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl; // Put the breakpoint here
        return;
    default:
        return;
    }
}

// Simple timer function for animations
void timer(int)
{
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    glutInitContextVersion(3, 3);
    glutInitContextProfile(GLUT_CORE_PROFILE);
    glutInitContextFlags(GLUT_DEBUG);

    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - test3");

    glewExperimental = GL_TRUE;
    glewInit();

    // Initialize DevIL library
    ilInit();

    SetDebugCallback(simple_debug_callback);

    init();

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutTimerFunc(20, timer, 0);
    glutMouseFunc(mouse_button_changed);
    glutMotionFunc(mouse_moved);

    glutMainLoop();

    return 0;
}
